var engine = function (eventAggregator, ScreenService, General, LibLoginController, ZhuAclController, LibUtilities, restEngine, requestDataSourceEngine, localStorageEngine, ApplicationContext, ZhUtil, ZHPromise) {

    //this.templateManager    = templateManager;
    this.localStorageEngine = localStorageEngine;
    var globPromise;
    var loadingOff = true;
    
    var aggregatorEvents = {
        BEFORE_PROJECT_STARTUP: "beforeProjectStartup",
        ON_CONTAINER_OPEN: "onContainerOpen",
        AUTH_SUCCESS: "authSuccess",
        LOGIN_SUCCESS: "loginSuccess"
    };
    var regularExpressions = {
        LETTERS: /((a|A))/g,
        LETTERS_AND_NUMBERS: /((a|A|9))/g
    };
    var messages = {
        userDontHaveUnits: "Usuário não possui unidades relacionadas",
        unsetGeneralParams: "Parâmetros gerais não configurados",
        logout: "Deseja realmente sair do sistema?"
    };
    var messageTypes = {
        ERROR: "E",
        ALERT: "W"
    };
    
    // Operador
    window.showFullOperatorWindow = true;
    window.filialRuleInFullOperatorWindow = true;

    eventAggregator.subscribe(aggregatorEvents.BEFORE_PROJECT_STARTUP, authUser);

    eventAggregator.subscribe(aggregatorEvents.AUTH_SUCCESS, onAuthSuccess.bind(this));

    function authUser() {
        if(typeof memoryStorage !== 'undefined') {
            localStorageEngine.getLocalVar = memoryStorage.getLocalVar;
            localStorageEngine.setLocalVar = memoryStorage.setLocalVar;
            setLocalVar = memoryStorage.setLocalVar;
            getLocalVar = memoryStorage.getLocalVar;
        }
        
        var loginDefer = ZHPromise.defer();
        localStorageEngine.setLocalVar('loginPromise', loginDefer.promise);
        globPromise = loginDefer;
        
        LibLoginController.authenticate()
            .then(validateLoginResponse)
            .catch(callbackError);
            
    }

    function onAuthSuccess(user) {
        setUserLocalVars(user);
        var userId = user.CDOPERADOR;
        if(ApplicationContext.ZhPreferences) {
            ApplicationContext.ZhPreferences.setUserID(userId);
        } else {
            ApplicationContext.zhCustomizationsService.registerUser(userId);
        }
        setGeneralParameters(user.nrorg).then(function () {
            if (userHasSelectedUnit(user)) {
                setLocalProcessDate(user.cdfilialselecionada).then(function () {
                    setHeaderInfoAndBuildUserData(user);
                }, callbackError);
                
                setCacheForTables();
            } else {
                openWindow('cadEst#ger01742_selecionar_unidade');
            }
        }, callbackError);
    }

    function validateLoginResponse(response) {
        var defaultResponse = response;
        response = LibUtilities.changeLoginResponseToOldParams(response);
        response.admin = defaultResponse.ACL_ADMINISTRATOR;
        response.token = defaultResponse.LOGIN_TOKEN;
        response.redirectUrl = defaultResponse.REDIRECT_URL;
        response.loginUserCrypto = defaultResponse.LOGIN_USER;
        
        if (response && response.cdoperador) {
            startSession(response).then(function(callback) {
                eventAggregator.publish(aggregatorEvents.LOGIN_SUCCESS, defaultResponse);
                eventAggregator.publish(aggregatorEvents.AUTH_SUCCESS, response);
                globPromise.resolve();
                startNext(response).then(function(callbackNext){
                    setLocalVar("NEXTSESSION_ENCRYPTED", callbackNext.dataset.NEXTSESSION_ENCRYPTED[0]);
                }.bind(this));
            }.bind(this));
        } 
    }
    
    var setHeaderInfoAndBuildUserData = function(response){
        var filter = [];
    
        requestDataSourceEngine.getDataSource('acl#/findUserData', filter).then(function(data) {
            if(typeof data == 'object') {
                var userData     = data.dataset.findUserData[0];
                //var defaultImage = window.modulesPackage ? OMConfig.defaultImageProduct : OMConfig.defaultImageModules;
                var userImage    = userData.IMAGE || '';
            }
            else userImage = '';
            
            var nmoperador = getLocalVar('NMOPERADOR');
            var dtprocessa = getLocalVar('DTPROCESSA');
            var cdfilial = getLocalVar('CDFILIAL');
            var nmfilial = getLocalVar('NMFILIAL');
            var unidade = cdfilial + " - " + nmfilial;
            var headerInfo = dtprocessa + " | " + unidade;
            ScreenService.buildUserData(nmoperador, [dtprocessa, unidade],
                function() {
                    LibUtilities.openLinkByMenu('myProfile');
                },
                userImage);
            
            ScreenService.setHeaderInfo(headerInfo);
        },
        function(error) {
            var nmoperador = getLocalVar('NMOPERADOR');
            var dtprocessa = getLocalVar('DTPROCESSA');
            var cdfilial = getLocalVar('CDFILIAL');
            var nmfilial = getLocalVar('NMFILIAL');
            var unidade = cdfilial + " - " + nmfilial;
            var headerInfo = dtprocessa + " | " + unidade;
            ScreenService.buildUserData(nmoperador, [dtprocessa, unidade],
                function() {
                    LibUtilities.openLinkByMenu('myProfile');
                },
                '');
            
            ScreenService.setHeaderInfo(headerInfo);
        });
    };

    function startSession(user) {
        return General.callRestEngine("product#/startSession", "Row", user, loadingOff).then(null, callbackError);
    }
    
    function startNext(user) {
        return General.callRestEngine("product#/startNext", "Row", user, loadingOff).then(null, callbackError);
    }

    function isInvalidHash() {
        return Util.isEmptyOrBlank(window.location.hash);
    }

    function isNewOperator(cdoperador) {
        return Util.isEmptyOrBlank(getLocalVar('CDOPERADOR')) || getLocalVar('CDOPERADOR') != cdoperador;
    }

    function userHasSelectedUnit(user) {
        return !Util.isEmptyOrBlank(user.cdfilialselecionada);
    }

    function setUserLocalVars(user) {
        var cdoperador = user.cdoperador;
        var perfilDeAcesso = user.perfilDeAcesso;
        var userAcl = {
            NAME:  user.nmoperador,
            EMAIL: user.dsemailoper
        };
        var accessProfile = [];
        user.perfilDeAcesso.forEach(function(perfil){
            accessProfile.push({
                ID: perfil,
                IDPERFILACESSOADM: user.admin
            });
        }.bind(this));
        
        //Zeedhi Access Control
        setLocalVar("userId", cdoperador);
        setLocalVar("aclUser", userAcl);
        setLocalVar("accessProfiles", accessProfile); 
        setLocalVar("productIdLoginUrl", 1078);
        
        setLocalVar('CDOPERADOR', user.cdoperador);
        setLocalVar('NMOPERADOR', user.nmoperador);
        setLocalVar('CDFILIAL', user.cdfilialselecionada);
        setLocalVar('ERP_CDFILIAL', user.cdfilialselecionada);
        setLocalVar('NMFILIAL', user.nmfilialselecionada);
        setLocalVar('NRORG', user.nrorg);
        setLocalVar('NRORGTRAB', user.nrorgtrab);
        setLocalVar('ERP_NRORG', user.nrorgtrab);
        setLocalVar('EMAIL', user.dsemailoper);
    }   
    
    function getLocalVar(name) {
        return localStorageEngine.getLocalVar(name);
    }
    
    function setLocalVar(name, value) {
        localStorageEngine.setLocalVar(name, value);
    }

    function replaceMask(value, regex) {
        return typeof (value) === 'string' ? value.replace(regex, '*') : value;
    }

    function setGeneralParameters(nrorg) {
        return getParamgeral().then(function (paramgeral) {
            if (paramgeral) {
                setLocalVar('IDUTILDTPROC', paramgeral.IDUTILDTPROC);
                setLocalVar('CAD_CDPICTPRAT', replaceMask(paramgeral.CDPICTPRAT, regularExpressions.LETTERS));
                setLocalVar('GER_CDPICTPROD', replaceMask(paramgeral.CDPICTPROD, regularExpressions.LETTERS));
                setLocalVar('CDPICTPROD', replaceMask(paramgeral.CDPICTPROD, regularExpressions.LETTERS));
                setLocalVar('CDPICTCCTB', replaceMask(paramgeral.CDPICTCCTB, regularExpressions.LETTERS));
                setLocalVar('GER_CDPICTPRAT', replaceMask(paramgeral.CDPICTPRAT, regularExpressions.LETTERS_AND_NUMBERS));
                setLocalVar('CDPICTPRAT', replaceMask(paramgeral.CDPICTPRAT, regularExpressions.LETTERS_AND_NUMBERS));
            } else {
                ScreenService.showMessage(messages.unsetGeneralParams, messageTypes.ERROR);
            }
        });
    }
    
    function setCacheForTables(){
        let tables = [];
        tables.forEach(table => {
            getDataLocal(table).then(result =>{
                if(result.length == 0){
                    filterPagination(table, 'est#/crud/'+table).then( result => {
                        return localStorageEngine.setDataSource('est#/'+table+'CACHE', result.dataset[table]);
                    });
                }
            });
        });
    }

    function getParamgeral() {
        return getTableByName("paramgeral").then(function(response) {
            var paramgeral = null;
            if (response.length)
                paramgeral = response[0];
            return paramgeral;
        });
    }
    
    function getDataLocal(dataSourceName, filter) {
        filter = filter || [];
        return localStorageEngine.getDataSource('est#/'+dataSourceName+'CACHE', null, filter, null, 999999999);
    }
    
    function filterPagination(datasetName, route, filter = [], page = 1, registers = {dataset:{}}) {
        if(!registers.dataset[datasetName])
            registers = {'dataset': {[datasetName]: []}};
        return rest(route, 'FilterData', filter, {
            "itemsPerPage": 500,
            "page": page
        }).then(function(result) {
            var registersResult = result.dataset[datasetName];
            registers.dataset[datasetName] = registers.dataset[datasetName].concat(registersResult);
            if (registersResult.length < 500) {
                return registers;
            } else {
                page++;
                return filterPagination(datasetName, route, filter, page, registers);
            }
        });
    }
    
    function rest(dataSourceName, requestType, filter, otherParams) {
        return restEngine.post(dataSourceName, Util.extend({
            "requestType": requestType,
            "filter": filter
        }, otherParams || {}));
    };

    function getTableByName(tableName, filter) {
        if (!filter) {
            filter = [];
        }
        return General.callRestEngine("product#/crud/" + tableName, "FilterData", filter, loadingOff).then(function (response) {
            var result = [];
            if (response.dataset[tableName])
                result = response.dataset[tableName];
            return result;
        });
    }

    function setLocalProcessDate(unitSelected) {
        var filter = [{
            name: "CDFILIAL",
            operator: "=",
            value: unitSelected
        }];
        var utilizaDtProc = getLocalVar('IDUTILDTPROC') == 'S';
  
        return getTableByName("paramfilial", filter).then(function (response) {
                var dtprocessa = moment().format("L");
                if(utilizaDtProc) {
                    if (response.length && response[0].DTPROCESSA) {
                        dtprocessa = moment(response[0].DTPROCESSA, "DD/MM/YYYY hh:mm:ss").format("L");
                    }
                }
                else {
                    dtprocessa = moment().format("L");
                }
                setLocalVar('DTPROCESSA', dtprocessa);
                return dtprocessa;
        });
        
    }

    function callbackError(err) {
        if (err.data && err.data.error) {
            ScreenService.showMessage(err.data.error, messageTypes.ERROR);
        } else {
            ScreenService.showMessage(JSON.stringify(err), messageTypes.ERROR);
        }
    }

    this.logout = function () {
        ScreenService.confirmMessage(messages.logout, messageTypes.ALERT, disconnect);
    };

    function disconnect() {
        removeLocalVars();
        LibLoginController.logout();
    }

    function removeLocalVars() {
        localforage.clear();
        localStorage.clear();
    }

    function buildUserAndHeader() {
        var nmoperador = getLocalVar('NMOPERADOR');
        var dtprocessa = getLocalVar('DTPROCESSA');
        var cdfilial = getLocalVar('CDFILIAL');
        var nmfilial = getLocalVar('NMFILIAL');
        var unidade = cdfilial + " - " + nmfilial;
        var headerInfo = dtprocessa + " | " + unidade;
        ScreenService.buildUserData(nmoperador, [dtprocessa, unidade], userMethod);
        ScreenService.setHeaderInfo(headerInfo);
    }

    function userMethod() {
        ScreenService.openWindow('ger00002_operador');
    }

    var requestModulesMetadata = function () {
		var modulesPackage = window.modulesPackage || "modules";
		return restEngine.requestMetaData(modulesPackage + "#json/modules.json").then(function (response) {
			return response.data;
		});
    };
    
    this.openWindow = openWindow;
    function openWindow(windowName, newWindow) {
        requestModulesMetadata().then(function (modules) {
            var subModule = windowName.split('#')[0];
            var module = subModule.split('_')[0];
            var url = window.location.origin + '/' + modules[module].baseUrl + "/#/" + windowName;
            if (ScreenService.openLink) {
                ScreenService.openLink(url, newWindow);
            } else {
                ScreenService.openWindow(windowName);
            }
        });
    };
    
    /*if(!ApplicationContext.setLocalVar) {
        var getLocalVar = function(name) {
    		var data = localStorage.getItem(name);
    		if (data && data !== 'undefined') {
    			return (JSON.parse(data));
    		} else {
    			return;
    		}
    	};
    	
    	var setLocalVar = function(name, localVar) {
    		var cleaned;
    		
    		var _clearArrayVar = function(localVar) {
        		var result = localVar;
        		if (Util.isObject(localVar)) {
        			result = ZhUtil.clearRow(localVar);
        		}
        		return result;
        	};
        	
    		if (Array.isArray(localVar)) {
    			cleaned = localVar.map(_clearArrayVar);
    		} else if (Util.isObject(localVar)) {
    			cleaned = ZhUtil.clearRow(localVar);
    		} else {
    			cleaned = localVar;
    		}
    		localStorage.setItem(name, JSON.stringify(cleaned));
    	};
    	//Para ser usado nos jsons com '@getLocalVar@'
    	ApplicationContext.registerInstance('setLocalVar', setLocalVar);
    	ApplicationContext.registerInstance('getLocalVar', getLocalVar);
    	//Para ser usado nos controllers com 'getLocalVar()'
    	window.setLocalVar = setLocalVar;
    	window.getLocalVar = getLocalVar;
    }*/
    
};

Configuration(function (ContextRegister) {
    ContextRegister.register('engine', engine);
});